from ..openai import Openai
from .schemas import OpenAIV1ChatCompletion
from fastapi import Request, APIRouter, status
from fastapi.responses import JSONResponse

router = APIRouter(prefix="/v1", tags=["v1"])
openai = Openai()


@router.get("/models")
async def list_models(request: Request):
    return await openai.v1_list_models(request)


@router.get("/models/{model}")
async def retrive_model(model: str, request: Request):
    return await openai.retrive_model(request, model)


@router.post("/chat/completions")
async def chat_completions(params: OpenAIV1ChatCompletion, request: Request):
    return await openai.v1_chat_completions(params, request)


@router.post("/completions")
async def completions(data, request: Request):
    """Given a prompt, the model will return one or more
    predicted completions, and can also return the probabilities
    of alternative tokens at each position.
    """
    return openai.v1_completions(data, request)


@router.post("/edits")
async def edits(data, request: Request):
    """Given a prompt and an instruction, the model will
    return an edited version of the prompt.
    """
    return openai.v1_edits(data, request)


@router.post("/images/generations")
async def image_generations(data, request: Request):
    """Creates an image given a prompt.
    """
    return openai.v1_image_generations(data, request)


@router.post("/embeddings")
async def embeddings(data, request: Request):
    """Given a prompt, the model will return a vector embedding
    of the prompt.
    """
    return openai.v1_embeddings(data, request)


# @router.post("/audio/transcriptions")
# async def audio_transcriptions(data, request: Request):
#     return openai.v1_audio_transcriptions(data, request)

@router.post("/fine-tunes")
async def fine_tunes(data, request: Request):
    return await openai.v1_fine_tunes(data, request)


@router.get("/fine-tunes")
async def fine_tunes(request: Request):
    return await openai.v1_fine_tunes(None, request)


@router.get("/fine-tunes/{fine_tune_id}")
async def fine_tunes(fine_tune_id, request: Request):
    return await openai.v1_retrieve_fine_tunes(fine_tune_id, request)


@router.get("/chat/completions")
def send_warning():
    return JSONResponse(status_code=status.HTTP_400_BAD_REQUEST, content={
        "error": {
            "message": "You didn't provide an API key. You need to provide your API key in an Authorization header using Bearer auth (i.e. Authorization: Bearer YOUR_KEY), or as the password field (with blank username) if you're accessing the API from your browser and are prompted for a username and password. You can obtain an API key from https://platform.openai.com/account/api-keys.",
            "type": "invalid_request_error",
            "param": None,
            "code": None
        }
    })



from .openai import Openai
from .routers.openai_v1 import router as router_v1
from sparrow.api import create_app
from fastapi import Request, APIRouter
import pretty_errors

Openai.stream_timeout = 18
Openai.non_stream_timeout = 30
openai = Openai()
app = create_app(title="openai_forward", version="1.0")
app.include_router(router_v1)


@app.get("/dashboard/billing/credit_grants", tags=['public'])
async def credit_grants(request: Request):
    return await openai.credit_grants(request)


@app.get("/dashboard/billing/usage", tags=['public'])
async def billing_usage(start_date: str, end_date: str, request: Request):
    params = {"start_date": start_date, "end_date": end_date}
    return await openai.billing_usage(params, request)



from __future__ import annotations

from pydantic import BaseModel, Field
from typing import List, Dict, Optional, Union


class OpenAIV1ChatCompletion(BaseModel):
    """Creates a completion for the chat message"""
    model: str = Field(..., description="The model to use for the completion", example="gpt-3.5-turbo")
    messages: List[Dict[str, str]] = Field(..., description="The message to complete",
                                           example=[{"role": "user", "content": "hi"}])
    temperature: float = Field(default=1, description="The temperature to use for the completion")
    top_p: float = Field(default=1, description="""An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10% probability mass are considered.\nWe generally recommend altering this or temperature but not both.""")
    n: int = Field(default=1, description="How many chat completion choices to generate for each input message.")
    stream: bool = Field(default=False)
    stop: Union[List[str], str, None] = Field(default=None,
                                              description="Up to 4 sequences where the API will stop generating further tokens.")
    max_tokens: Union[int, None] = Field(default=None, description="The maximum number of tokens to generate in the chat completion. The total length of input tokens and generated tokens is limited by the model's context length.")
    presence_penalty: float = Field(default=0, description="Number between -2.0 and 2.0. ")
    frequency_penalty: float = Field(default=0, description="Number between -2.0 and 2.0. ")
    logit_bias: Optional[Dict[str, float]] = Field(default=None, description="A dictionary of token to bias. ")
    user: Optional[str] = Field(default=None,
                                description="A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse. ",
                                )

    class Config:
        schema_extra = {
            "example": {
                "model": "gpt-3.5-turbo",
                "messages": [{"role": "user", "content": "hi"}],
                "stream": False
            }
        }



from ._base import OpenaiBase, Request
from .routers.schemas import OpenAIV1ChatCompletion
from dotenv import load_dotenv
import os

load_dotenv()


class Openai(OpenaiBase):
    def __init__(self):
        self.defualt_auth = os.environ.get("OPENAI_API_KEY", "")

    async def _forward(self, route: str, request: Request, data=None, validate_host=False):
        url = os.path.join(self.base_url, route)
        if validate_host:
            self.validate_request_host(request.client.host)
        return await self.forwarding(url,
                                     request,
                                     data,
                                     default_openai_auth=self.defualt_auth)

    async def credit_grants(self, request: Request):
        url = os.path.join(self.base_url, "dashboard/billing/credit_grants")
        return await self.forwarding(url, request=request,
                                     default_openai_auth=self.defualt_auth)

    async def billing_usage(self, params: dict, request: Request):
        url = os.path.join(self.base_url, "dashboard/billing/usage")
        return await self.forwarding(url, request=request, params=params, non_stream_timeout=5,
                                     default_openai_auth=self.defualt_auth)

    async def v1_chat_completions(self, data: OpenAIV1ChatCompletion, request: Request, validate_host=False):
        return await self._forward("v1/chat/completions", request, data, validate_host)

    async def v1_list_models(self, request: Request, validate_host=False):
        return await self._forward("v1/models", request, data=None, validate_host=validate_host)

    async def retrive_model(self, request: Request, model: str, validate_host=False):
        return await self._forward(f"v1/models/{model}", request, data=None, validate_host=validate_host)

    async def v1_completions(self, data, request: Request, validate_host=False):
        return await self._forward("v1/completions", request, data=data, validate_host=validate_host)

    async def v1_edits(self, data, request: Request, validate_host=False):
        return await self._forward("v1/edits", request, data=data, validate_host=validate_host)

    async def v1_image_generations(self, data, request: Request, validate_host=False):
        return await self._forward("v1/images/generations", request, data=data, validate_host=validate_host)

    async def v1_embeddings(self, data, request: Request, validate_host=False):
        return await self._forward("v1/embeddings", request, data=data, validate_host=validate_host)

    # async def v1_audio_transcriptions(self, data, request: Request, validate_host=False):
    #     return await self._forward("audio/transcriptions", request, data=data, validate_host=validate_host)

    async def v1_fine_tunes(self, data, request: Request, validate_host=False):
        return await self._forward("v1/fine-tunes", request, data=data, validate_host=validate_host)

    async def v1_retrieve_fine_tunes(self, path_var, request: Request, validate_host=False):
        return await self._forward(f"v1/fine-tunes/{path_var}", request, validate_host=validate_host)



import fire
import uvicorn


class Cli:
    def run(self, port=8000, workers=1):
        uvicorn.run(
            app="openai_forward.app:app",
            host="0.0.0.0",
            port=port,
            workers=workers,
            app_dir='..'
        )

def main():
    fire.Fire(Cli)



from loguru import logger
from sparrow import relp
import sys
import logging


class InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def setting_log(log_name, multi_process=True):
    logging.root.handlers = [InterceptHandler()]
    for name in logging.root.manager.loggerDict.keys():
        logging.getLogger(name).handlers = []
        logging.getLogger(name).propagate = True
    logger_config = {
        "handlers": [
            {"sink": sys.stdout},
            {"sink": relp(f"../Log/{log_name}"), "enqueue": multi_process, "rotation": "100 MB"},
        ],
    }
    logger.configure(**logger_config)



from fastapi import Request, Response, HTTPException, status
from fastapi.responses import StreamingResponse, RedirectResponse, FileResponse, JSONResponse
import requests
from loguru import logger
from .config import setting_log

setting_log(log_name="openai_forward.log")


class OpenaiBase:
    base_url = "https://api.openai.com"
    stream_timeout = 18
    non_stream_timeout = 30
    allow_ips = []

    def add_allowed_ip(self, ip: str):
        if ip == "*":
            ...
        else:
            self.allow_ips.append(ip)

    def validate_request_host(self, ip):
        if ip == "*" or ip in self.allow_ips:
            return True
        else:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                                detail=f"Forbidden, please add {ip=} to `allow_ips`")

    @staticmethod
    def try_get_response(url, method, headers, params, payload, stream, timeout):
        if params is None:
            params = {}
        def _exec(time_out):
            if method == 'post':
                return requests.post(url, headers=headers, params=params, json=payload, stream=stream,
                                     timeout=time_out)
            elif method == 'get':
                return requests.get(url, params=params, headers=headers, timeout=time_out)
            else:
                logger.error(f"method {method} not supported")
                raise NotImplementedError
        if stream:
            for i, current_timeout in enumerate([1.5, timeout, 2.5, 2.5, 1.5]):
                try:
                    logger.debug(f"try {i+1} times, timeout={current_timeout}")
                    return _exec(time_out=current_timeout)
                except:
                    ...
            raise HTTPException(status_code=status.HTTP_408_REQUEST_TIMEOUT, detail="Request Timeout")
        else:
            return _exec(time_out=timeout)

    async def forwarding(self, url, request: Request, params=None, data=None, default_openai_auth=None,
                         non_stream_timeout=None):
        non_stream_timeout = non_stream_timeout or self.non_stream_timeout
        method = request.method.lower()
        if data is not None:
            payload_tmp = data.dict()
            payload = payload_tmp.copy()
            for key, value in payload_tmp.items():
                if value is None:
                    payload.pop(key)
        else:
            try:
                payload = await request.json()
            except:
                payload = {}
        stream = payload.get('stream')
        timeout = self.stream_timeout if stream else non_stream_timeout

        geted_headers = dict(request.headers)
        posted_auth = geted_headers.get("authorization")
        if posted_auth and str(posted_auth).startswith("Bearer sk-"):
            auth = posted_auth
        else:
            if default_openai_auth:
                auth = "Bearer " + default_openai_auth
            else:
                return JSONResponse(status_code=status.HTTP_400_BAD_REQUEST, content={
                    "error": {
                        "message": "You didn't provide an API key. You need to provide your API key in an Authorization header using Bearer auth (i.e. Authorization: Bearer YOUR_KEY), or as the password field (with blank username) if you're accessing the API from your browser and are prompted for a username and password. You can obtain an API key from https://platform.openai.com/account/api-keys.",
                        "type": "invalid_request_error",
                        "param": None,
                        "code": None
                    }
                })
        headers = {
            "Content-Type": "application/json",
            "Authorization": auth
        }
        logger.debug(f"{payload.get('messages')=}")
        # logger.debug(f"{headers=}")
        response = self.try_get_response(url=url, method=method,
                                         params=params,
                                         headers=headers, payload=payload, stream=stream,
                                         timeout=timeout)
        if response.status_code != 200:
            raise HTTPException(status_code=response.status_code, detail=response.text)
        if stream:
            return StreamingResponse(content=response.iter_content(chunk_size=32),
                                     status_code=response.status_code,
                                     media_type=response.headers.get("content-type"))
        else:
            return Response(content=response.content,
                            status_code=response.status_code,
                            media_type=response.headers.get("content-type"))
